#if !defined(CURRENT_GPU)
#  include "../tuning/CURRENT_GPU"
#endif
#if !defined(GPU_ARCH)
#  define GPU_ARCH	CURRENT_GPU
#endif

#include "float_complex.h"
#include "aspen_devel.h"

#if ( 'c'!='h' && 'c'!='k' ) || ASPEN_HALF_ENABLED
#  include "chemv_upper-X_template.cu"
#endif

extern "C" void
API(CHEMVu) ( const int blk,
              const int n, const scalar_t * a, const int lda,
              const scalar_t * x, const int incx,
              scalar_t * y, const int incy,
              const scalar_t alpha, const scalar_t beta )
{

  // y = alpha * A * x + beta * y
  // y(1:n1)
  // x(1:n2)

#if ( 'c'!='h' && 'c'!='k' ) || ASPEN_HALF_ENABLED

  int Ver = API(get_device_CC) ( );

  if ( CURRENT_GPU == Ver ) {
    API_private(CHEMVu_STUB) ( blk,
                               n, a, lda, x, incx, y, incy, alpha, beta );
  } else {
    perror( "unmatched GPU generation" ); /* unmatched */
  }

#endif

}

#if ( 'c'!='h' && 'c'!='k' ) || ASPEN_HALF_ENABLED
#  include "aspen_devel_postdef.h"
#  include "../template/trmv_n_upper_for_symv_template.incl"
#  include "../template/trmv_t_upper_for_symv_template.incl"
#endif

extern "C"
void
chemv_upper_small (
              const int n,
              const scalar_t alpha,
              const scalar_t * a, const int lda,
              const scalar_t * x, const int incx,
              const scalar_t beta,
              _____ scalar_t * y, const int incy
        )
{
#if ( 'c'!='h' && 'c'!='k' ) || ASPEN_HALF_ENABLED

  if ( n <= 0 ) return;

  API(TRMV_n_upper_for_SYMV_host)
#if CURRENT_GPU==800
    < scalar_t, 16, 32 >
#else
#if ( 'c'=='d' || 'c'=='z' || 'c'=='w' || 'c'=='u' || ( 'c'=='i' && 8>=8 ) )
    < scalar_t, 8, 40 >
#else
    < scalar_t, 16, 20 >
#endif
#endif
    ( n, alpha, beta, a, lda, x, incx, y, incy );
  API(TRMV_t_upper_for_SYMV_host)
#if CURRENT_GPU==800
    < scalar_t, 128, 1 >
#else
#if ( 'c'=='d' || 'c'=='z' || 'c'=='w' || 'c'=='u' || ( 'c'=='i' && 8>=8 ) )
    < scalar_t, 16, 6 >
#else
    < scalar_t, 32, 8 >
#endif
#endif
    ( n, alpha, beta, a, lda, x, incx, y, incy );

#endif
}

